@@ -21,6 +21,7 @@ class OperatingSystemFamily(Enum):
     DEBIAN = "debian"
     REDHAT = "redhat"
     SUSE = "suse"
+    ARCHLINUX = "archlinux"
 
 
 class OperatingSystemName(Enum):
@@ -28,9 +29,11 @@ class OperatingSystemName(Enum):
     UBUNTU = "ubuntu"
     FEDORA = "fedora"
     CENTOS = "centos"
+    RHEL = "rhel"
     ROCKYLINUX = "rockylinux"
     ARCHLINUX = "archlinux"
     OPENSUSE = "opensuse"
+    SLES = "sles"
     AMAZONLINUX = "amazonlinux"
     ORACLELINUX = "oraclelinux"
 
